import copy
import csv
import math
import os
from collections import defaultdict, OrderedDict
from typing import List

import matplotlib

matplotlib.use("Qt5Agg")

import matplotlib.pyplot as plt

from decide.model.base import Issue
from .. import base
from .. import calculations
from ..observers import observer


class IssueDevelopment(observer.Observer):
    """
    There are three stages of externalities
    By exchange, by issue set and by actor
    """

    def __init__(
        self,
        observable: observer.Observable,
        write_voting_position=False,
        summary_only=False,
    ):
        super().__init__(observable=observable)

        self.preference_history = {}
        self.voting_history = {}
        self.voting_loss = {}
        self.preference_loss = {}
        self.issue_obj = None
        self.write_voting_position = write_voting_position

        self.preference_history_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.voting_history_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.voting_loss_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.preference_loss_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )

        self.denominator = 0

        self.summary_only = summary_only

    def _setup(self):
        """
        Setup method.
        """
        for issue in self.model_ref.actor_issues:
            issue_list = {}

            for key, actor_issue in self.model_ref.actor_issues[issue].items():
                issue_list[actor_issue.actor] = []

            issue_list["nbs"] = []

            self.preference_history[issue] = copy.deepcopy(issue_list)
            self.voting_history[issue] = copy.deepcopy(issue_list)
            self.voting_loss[issue] = copy.deepcopy(issue_list)
            self.preference_loss[issue] = copy.deepcopy(issue_list)

        self.denominator += 1

    def _de_normalize_value(self, value):
        return self.issue_obj.de_normalize(value)

    def _de_normalize_list_value(self, values, _map=None):
        for _, value in enumerate(values):
            if _map is not None:
                values[_] = self.issue_obj.de_normalize(_map(value))
            else:
                values[_] = self.issue_obj.de_normalize(value)
        return values

    def _create_directories(self, repetition):
        """
        Create the directories
        :param repetition:
        :return:
        """
        if not os.path.exists(
            "{0}/issues/{1}".format(self.output_directory, repetition)
        ):
            os.makedirs("{0}/issues/{1}/csv".format(self.output_directory, repetition))
            os.makedirs(
                "{0}/issues/{1}/charts".format(self.output_directory, repetition)
            )

    def before_repetitions(self, repetitions, iterations, randomized_value=None):
        pass

    def before_iterations(self, repetition):
        """
        :param repetition:
        :return:
        """
        self._setup()

        if not self.summary_only:
            self._create_directories(repetition)

    def before_loop(self, iteration: int, repetition: int):
        pass

    def after_loop(
        self, realized: List[base.AbstractExchange], iteration: int, repetition: int
    ):
        """
        After each round we calculate the variance
        :param repetition:
        :param realized:
        :param iteration:
        :return:
        """
        model = self.model_ref

        for issue in model.actor_issues:

            nbs = model.nbs[issue]
            variance_sum = 0

            for actor_id, actor_issue in model.actor_issues[issue].items():
                # only the sum part of the variance formula
                variance_sum += (actor_issue.position - nbs) ** 2

                # the distance of the actor adjusted by his salience
                actor_loss = abs(actor_issue.position - nbs) * actor_issue.salience

                # track history
                self.preference_history[issue][actor_id].append(actor_issue.position)
                self.preference_history_sum[issue][actor_id][iteration].append(
                    actor_issue.position
                )

                # track loss
                self.preference_loss[issue][actor_id].append(actor_loss)
                self.preference_loss_sum[issue][actor_id][iteration].append(actor_loss)

            nbs_var = variance_sum / len(model.actor_issues[issue])

            self._append_nbs_preferences(
                issue=issue, iteration=iteration, nbs=nbs, nbs_var=nbs_var
            )

    def _append_nbs_preferences(self, issue, iteration, nbs, nbs_var):
        self.preference_history[issue]["nbs"].append(nbs)
        self.preference_history_sum[issue]["nbs"][iteration].append(nbs)
        self.preference_loss[issue]["nbs"].append(nbs_var)
        self.preference_loss_sum[issue]["nbs"][iteration].append(nbs_var)

    def end_loop(self, iteration: int, repetition: int):
        """
        Before each round we calculate the voting position
        :param repetition:
        :param iteration:1
        :return:
        """

        for issue in self.model_ref.actor_issues:

            nbs = self.model_ref.nbs[issue]
            variance_sum = 0

            for actor_id, actor_issue in self.model_ref.actor_issues[issue].items():
                variance_sum += (actor_issue.position - nbs) ** 2

                self.voting_history[issue][actor_id].append(actor_issue.position)
                self.voting_loss[issue][actor_id].append(
                    abs(actor_issue.position - nbs) * actor_issue.salience
                )

            nbs_var = variance_sum / len(self.model_ref.actor_issues[issue])

            self.voting_history[issue]["nbs"].append(nbs)
            self.voting_loss[issue]["nbs"].append(nbs_var)

    def after_iterations(self, repetition):
        """
        Write all the data of this repetition to the filesystem
        """

        if self.summary_only:
            return

        for issue in self.preference_history:
            with open(
                "{0}/issues/{2}/csv/{1}.{3}.csv".format(
                    self.output_directory, issue, repetition, self._get_salt
                ),
                "w",
            ) as csv_file:
                writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")

                self.issue_obj = self.model_ref.issues[issue]  # type: Issue

                heading = [
                    "rnd-" + str(x)
                    for x in range(len(self.preference_history[issue]["nbs"]))
                ]

                writer.writerow([self.issue_obj])
                writer.writerow([])  # empty row for readability

                writer.writerow(["Issue overview"])

                if self.write_voting_position:
                    writer.writerow(
                        [
                            "round",
                            "preference nbs",
                            "preference nbs var",
                            "",
                            "voting nbs",
                            "voting nbs var",
                        ]
                    )
                else:
                    writer.writerow(["round", "preference nbs", "preference nbs var"])

                # remove the nbs from each collection so it doesn't show up in the table.
                preference_nbs = self.preference_history[issue]["nbs"]
                del self.preference_history[issue]["nbs"]

                preference_nbs_var = self.preference_loss[issue]["nbs"]
                del self.preference_loss[issue]["nbs"]

                if self.write_voting_position:
                    voting_nbs = self.voting_history[issue]["nbs"]
                    del self.voting_history[issue]["nbs"]

                    voting_nbs_var = self.voting_loss[issue]["nbs"]
                    del self.voting_loss[issue]["nbs"]

                # create a summary table for only the preference and voting nbs + variance
                for x in range(len(preference_nbs)):

                    row = [
                        "rnd-" + str(x),  # round
                        self._de_normalize_value(preference_nbs[x]),  # preference nbs
                        self._de_normalize_value(
                            preference_nbs_var[x]
                        ),  # preference nbs variance
                    ]

                    # write the voting positions only when the author asks for it te keep the output simple.
                    if self.write_voting_position:
                        row.append("")  # extra spacing for readability
                        row.append(
                            self._de_normalize_value(voting_nbs[x])
                        )  # voting nbs
                        row.append(
                            self._de_normalize_value(voting_nbs_var[x])
                        )  # voting nbs variance

                    writer.writerow(row)

                # issue development summary
                writer.writerow([])
                writer.writerow(
                    ["First and last round comparison of NBS and all actors"]
                )
                writer.writerow(
                    [
                        "Actor",
                        "Salience",
                        "Power",
                        "First round",
                        "Final round",
                        "Position shift",
                        "Distance NBS start",
                        "Distance NBS end",
                    ]
                )

                nbs_start = self._de_normalize_value(preference_nbs[0])
                nbs_end = self._de_normalize_value(preference_nbs[-1])

                writer.writerow(["NBS", "-", "-", nbs_start, nbs_end, "-"])

                # to compare different issues and variances of the model, keep the output sorted
                od = OrderedDict(sorted(self.preference_history[issue].items()))

                for actor_id, value in od.items():
                    actor_issue = self.model_ref.actor_issues[self.issue_obj][actor_id]

                    position_start = self._de_normalize_value(value[0])
                    position_end = self._de_normalize_value(value[-1])
                    position_delta = position_end - position_start

                    nbs_distance_start = abs(position_start - nbs_start)
                    nbs_distance_end = abs(position_start - nbs_end)
                    nbs_distance_delta = nbs_distance_end - nbs_distance_start

                    writer.writerow(
                        [
                            actor_issue.actor,
                            actor_issue.salience,
                            actor_issue.power,
                            position_start,
                            position_end,
                            position_delta,
                            nbs_distance_start,
                            nbs_distance_end,
                            nbs_distance_delta,
                        ]
                    )

                # second table
                writer.writerow([])
                writer.writerow(["Preference development NBS and all actors"])
                writer.writerow(["actor", "salience", "power"] + heading)

                plt.clf()

                nbs_values = self._de_normalize_list_value(preference_nbs)

                writer.writerow(["nbs", "-", "-"] + nbs_values)

                plt.plot(nbs_values, label="nbs")

                for actor_id, value in od.items():
                    actor_issue = self.model_ref.actor_issues[self.issue_obj][actor_id]

                    values = self._de_normalize_list_value(value)
                    writer.writerow(
                        [actor_id, actor_issue.salience, actor_issue.power] + values
                    )

                    plt.plot(values, label=self.model_ref.actors[actor_id].name)

                lgd = plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
                plt.title(self.issue_obj)
                plt.savefig(
                    "{0}/issues/{2}/charts/{1}.png".format(
                        self.output_directory, self.issue_obj.name, repetition
                    ),
                    bbox_extra_artists=(lgd,),
                    bbox_inches="tight",
                )

                if self.write_voting_position:
                    writer.writerow([])
                    writer.writerow(["Voting development NBS and all actors"])
                    writer.writerow(["Actor", "Salience", "Power"] + heading)

                    writer.writerow(
                        ["NBS", "-", "-"] + self._de_normalize_list_value(voting_nbs)
                    )

                    od = OrderedDict(sorted(self.voting_history[issue].items()))

                    for actor_id, value in od.items():
                        actor_issue = self.model_ref.actor_issues[self.issue_obj][
                            actor_id
                        ]
                        writer.writerow(
                            [actor_id, actor_issue.salience, actor_issue.power]
                            + self._de_normalize_list_value(value)
                        )

                writer.writerow([])
                writer.writerow(["Preference variance and loss of all actors"])

                writer.writerow(["nbs-var"] + preference_nbs_var)

                od = OrderedDict(sorted(self.preference_loss[issue].items()))

                for actor_id, value in od.items():
                    writer.writerow([actor_id] + value)

                if self.write_voting_position:

                    writer.writerow([])
                    writer.writerow(["Voting variance and loss of all actors"])

                    writer.writerow(["nbs-var"] + voting_nbs_var)

                    od = OrderedDict(sorted(self.voting_loss[issue].items()))

                    for actor_id, value in od.items():
                        writer.writerow([actor_id] + value)

    def after_repetitions(self):
        """
        Write all the results to a summary with the averages for all repetitions
        :return:
        """

        self._create_directories("summary")

        for issue in self.preference_history_sum:
            with open(
                "{0}/issues/{2}/csv/{1}.{3}.{2}.csv".format(
                    self.output_directory, issue, "summary", self._get_salt
                ),
                "w",
            ) as csv_file:
                writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")

                self.issue_obj = self.model_ref.issues[issue]  # type: Issue

                heading = [
                    "rnd-" + str(x)
                    for x in range(len(self.preference_history_sum[issue]["nbs"]))
                ]

                writer.writerow([self.issue_obj])
                writer.writerow([])
                writer.writerow(["Overview issue"])

                if self.write_voting_position:
                    writer.writerow(
                        [
                            "round",
                            "avg nbs",
                            "variance",
                            "",
                            "avg voting nbs",
                            "variance",
                            "",
                            "avg nbs var",
                            "variance",
                            "",
                            "avg voting variance",
                            "variance",
                        ]
                    )
                else:
                    writer.writerow(
                        [
                            "round",
                            "avg nbs",
                            "variance",
                            "",
                            "avg voting nbs",
                            "variance",
                        ]
                    )

                preference_nbs = self.preference_history_sum[issue]["nbs"]
                del self.preference_history_sum[issue]["nbs"]

                voting_nbs = self.voting_history_sum[issue]["nbs"]
                del self.voting_history_sum[issue]["nbs"]

                preference_nbs_var = self.preference_loss_sum[issue]["nbs"]
                del self.preference_loss_sum[issue]["nbs"]

                voting_nbs_var = self.voting_loss_sum[issue]["nbs"]
                del self.voting_loss_sum[issue]["nbs"]

                _ = self._de_normalize_value

                # NBS related data.
                plt.clf()

                p_line = []

                nbs_start = None
                nbs_end = None

                for x in range(len(preference_nbs)):

                    if self.write_voting_position:

                        p = calculations.average_and_variance(preference_nbs[x])
                        v = calculations.average_and_variance(voting_nbs[x])
                        pvar = calculations.average_and_variance(preference_nbs_var[x])
                        vvar = calculations.average_and_variance(voting_nbs_var[x])

                        writer.writerow(
                            [
                                "rn-" + str(x),
                                _(p[0]),
                                p[1],
                                "",
                                _(v[0]),
                                v[1],
                                "",
                                _(pvar[0]),
                                pvar[1],
                                "",
                                _(vvar[0]),
                                vvar[1],
                            ]
                        )

                        p_line.append(_(p[0]))
                    else:
                        p = calculations.average_and_variance(preference_nbs[x])
                        pvar = calculations.average_and_variance(preference_nbs_var[x])

                        writer.writerow(
                            ["rn-" + str(x), _(p[0]), p[1], "", _(pvar[0]), pvar[1]]
                        )

                        p_line.append(_(p[0]))

                    if nbs_start is None:
                        nbs_start = _(p[0])
                    nbs_end = _(p[0])

                plt.plot(p_line, label="nbs")

                writer.writerow([])
                writer.writerow(
                    ["First and last round comparison of NBS and all actors"]
                )
                writer.writerow(
                    [
                        "Actor",
                        "Salience",
                        "Power",
                        "First round",
                        "Final round",
                        "Position shift",
                        "Distance NBS start",
                        "Distance NBS end",
                    ]
                )

                writer.writerow(["NBS", "-", "-", nbs_start, nbs_end, "-"])

                for actor, value in sorted(self.preference_history_sum[issue].items()):

                    actor_issue = self.model_ref.actor_issues[issue][actor]

                    first = None
                    last = 0

                    for iteration, values in value.items():
                        avg, var = calculations.average_and_variance(values)

                        if not first:
                            first = avg

                        last = avg

                    position_start = _(first)
                    position_end = _(last)
                    position_delta = position_end - position_start

                    nbs_distance_start = abs(position_start - nbs_start)
                    nbs_distance_end = abs(position_start - nbs_end)

                    writer.writerow(
                        [
                            actor_issue.actor,
                            actor_issue.salience,
                            actor_issue.power,
                            position_start,
                            position_end,
                            position_delta,
                            nbs_distance_start,
                            nbs_distance_end,
                        ]
                    )

                writer.writerow([])
                writer.writerow(["AVG Preference development NBS and all actors"])
                writer.writerow(["actor", "salience", "power"] + heading)

                var_rows = []

                for actor, value in sorted(self.preference_history_sum[issue].items()):

                    actor_issue = self.model_ref.actor_issues[issue][actor]

                    row = [actor, actor_issue.salience, actor_issue.power]
                    var_row = [actor, actor_issue.salience, actor_issue.power]

                    avg_row = []
                    for iteration, values in value.items():
                        avg, var = calculations.average_and_variance(values)
                        avg_row.append(_(avg))
                        var_row.append(math.sqrt(var))

                    plt.plot(avg_row, label=actor.name)

                    writer.writerow(row + avg_row)
                    var_rows.append(var_row)

                writer.writerow([])
                writer.writerow(["Standard deviation"])
                writer.writerow(["actor", "salience", "power"] + heading)

                for row in var_rows:
                    writer.writerow(row)

                lgd = plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
                plt.title(self.issue_obj)
                plt.savefig(
                    "{0}/issues/{2}/charts/{1}.png".format(
                        self.output_directory, self.issue_obj.issue_id, "summary"
                    ),
                    bbox_extra_artists=(lgd,),
                    bbox_inches="tight",
                )

    @property
    def _get_salt(self):
        model_name = "random"
        from decide.model.equalgain import EqualGainModel

        if isinstance(self.model_ref, EqualGainModel):
            model_name = "equal"
            if self.model_ref.randomized_value is not None:
                model_name += "-" + str(round(self.model_ref.randomized_value, 2))

        return model_name



import datetime

import peewee
from playhouse.db_url import connect

connection = peewee.DatabaseProxy()


class DictionaryIndexMixin:
    hash_field = "id"

    def __init__(self, *args, **kwargs):
        super(DictionaryIndexMixin, self).__init__(*args, **kwargs)

    def __hash__(self):
        return hash(self.get_hash_field())

    def __eq__(self, other):

        if hasattr(other, "get_hash_field"):
            return self.get_hash_field() == other.get_hash_field()

        if self.get_hash_field() == other:
            return True

        if other is None:
            return False

        raise NotImplementedError()

    def get_hash_field(self):
        return getattr(self, self.__class__.hash_field)


class BaseModel(peewee.Model):
    class Meta:
        database = connection


class DataSet(DictionaryIndexMixin, BaseModel):
    name = peewee.CharField()
    hash_field = "name"


class Actor(DictionaryIndexMixin, BaseModel):
    name = peewee.CharField()
    key = peewee.CharField()

    data_set = peewee.ForeignKeyField(DataSet)

    hash_field = "key"


class Issue(DictionaryIndexMixin, BaseModel):
    name = peewee.CharField()
    key = peewee.CharField()

    lower = peewee.DecimalField()
    upper = peewee.DecimalField()

    data_set = peewee.ForeignKeyField(DataSet, on_delete="CASCADE")

    hash_field = "key"


class ModelRun(BaseModel):
    started_at = peewee.DateTimeField(default=datetime.datetime.now)
    finished_at = peewee.DateTimeField(null=True)

    p = peewee.DecimalField(max_digits=3, decimal_places=2)
    iterations = peewee.IntegerField()
    repetitions = peewee.IntegerField()

    data_set = peewee.ForeignKeyField(DataSet, on_delete="CASCADE")


class Repetition(DictionaryIndexMixin, BaseModel):
    """
    A repetition for a data set
    """

    pointer = peewee.IntegerField()

    hash_field = "pointer"

    model_run = peewee.ForeignKeyField(ModelRun, on_delete="CASCADE")


class Iteration(DictionaryIndexMixin, BaseModel):
    """
    A iteration inside a repetition
    """

    pointer = peewee.IntegerField()
    repetition = peewee.ForeignKeyField(Repetition, on_delete="CASCADE")

    hash_field = "pointer"


class ActorIssue(BaseModel):
    """
    Snapshot of a position of an actor on an issue
    """

    issue = peewee.ForeignKeyField(Issue, on_delete="CASCADE")

    actor = peewee.ForeignKeyField(Actor, on_delete="CASCADE")

    power = peewee.DecimalField(max_digits=20, decimal_places=15)
    position = peewee.DecimalField(max_digits=20, decimal_places=15)
    salience = peewee.DecimalField(max_digits=20, decimal_places=15)

    iteration = peewee.ForeignKeyField(Iteration, on_delete="CASCADE")

    type = peewee.CharField(choices=("before", "after"), default="before")

    def __str__(self):
        return "{issue} {actor} {position}".format(
            issue=self.issue.name, actor=self.actor.name, position=self.position
        )


class ExchangeActor(BaseModel):
    actor = peewee.ForeignKeyField(Actor, on_delete="CASCADE")
    supply_issue = peewee.ForeignKeyField(Issue, on_delete="CASCADE")

    demand_issue = peewee.ForeignKeyField(Issue, on_delete="CASCADE")

    x = peewee.DecimalField(max_digits=20, decimal_places=15)  # begin position
    y = peewee.DecimalField(max_digits=20, decimal_places=15)  # end position
    eu = peewee.DecimalField(
        max_digits=20, decimal_places=15
    )  # expected utility or gain

    demand_position = peewee.DecimalField(max_digits=20, decimal_places=15)

    # shortcut
    other_actor = peewee.ForeignKeyField("self", null=True, on_delete="CASCADE")

    @property
    def move(self):
        return abs(self.x - self.y)


class Exchange(BaseModel):
    i = peewee.ForeignKeyField(ExchangeActor, on_delete="CASCADE")

    j = peewee.ForeignKeyField(ExchangeActor, on_delete="CASCADE")

    iteration = peewee.ForeignKeyField(Iteration, on_delete="CASCADE")


class Externality(BaseModel):
    actor = peewee.ForeignKeyField(Actor, on_delete="CASCADE")

    exchange = peewee.ForeignKeyField(Exchange, on_delete="CASCADE")

    supply = peewee.ForeignKeyField(Issue, on_delete="CASCADE")
    demand = peewee.ForeignKeyField(Issue, on_delete="CASCADE")

    own = peewee.DecimalField(max_digits=20, decimal_places=15, null=True)
    inner_positive = peewee.DecimalField(max_digits=20, decimal_places=15, null=True)
    inner_negative = peewee.DecimalField(max_digits=20, decimal_places=15, null=True)
    outer_positive = peewee.DecimalField(max_digits=20, decimal_places=15, null=True)
    outer_negative = peewee.DecimalField(max_digits=20, decimal_places=15, null=True)

    iteration = peewee.ForeignKeyField(Iteration)


class Manager:
    """
    Helper for manage the state of the database.
    """

    tables = [
        DataSet,
        Actor,
        Issue,
        ModelRun,
        ActorIssue,
        Repetition,
        Iteration,
        Exchange,
        ExchangeActor,
        Externality,
    ]

    def __init__(self, database_path):
        self.database_path = database_path

    def init_database(self):
        global connection
        db = connect(self.database_path)
        connection.initialize(db)

    def create_tables(self):
        connection.create_tables(self.tables, safe=True)

    def delete_tables(self):
        connection.drop_tables(self.tables)

    def __call__(self):
        """
        Single instance only
        """
        return self



from blinker import Namespace

namespace = Namespace()

actor_created = namespace.signal("actor_created")
actor_changed = namespace.signal("actor_changed")
actor_deleted = namespace.signal("actor_deleted")

issue_created = namespace.signal("issue_created")
issue_changed = namespace.signal("issue_changed")
issue_deleted = namespace.signal("issue_deleted")

actor_issue_created = namespace.signal("actor_issue_created")
actor_issue_changed = namespace.signal("actor_issue_changed")
actor_issue_deleted = namespace.signal("actor_issue_deleted")



from typing import List, Dict

from decide.data import types
from decide.data.reader import InputDataFile
from decide.model.base import AbstractModel


class ModelFactory:
    """
    The goal of this object is to initiate a model object with the correct begin state
    """

    def __init__(
            self,
            date_file: InputDataFile,
            actor_whitelist: List[str] = None,
            issue_whitelist: List[str] = None,
            *args,
            **kwargs
    ):
        super(ModelFactory, self).__init__(*args, **kwargs)
        self.data_file = date_file
        self.actor_whitelist = actor_whitelist
        self.issue_whitelist = issue_whitelist

    def filter_actors(self) -> Dict[str, types.PartialActor]:

        if not self.actor_whitelist or len(self.actor_whitelist) == 0:
            return self.data_file.actors

        return {x: y for x, y in self.data_file.actors.items() if y.id in self.actor_whitelist}

    def filter_issues(self) -> Dict[str, types.PartialIssue]:

        if not self.issue_whitelist or len(self.issue_whitelist) == 0:
            return self.data_file.issues

        return {x: y for x, y in self.data_file.issues.items() if y.name in self.issue_whitelist}

    def filter_actor_issues(self, actors: Dict[str, types.PartialActor], issues: Dict[str, types.PartialIssue]):

        actor_issues = []

        for actor_issue in self.data_file.actor_issues.values():

            if actor_issue.actor in actors and actor_issue.issue in issues:
                actor_issues.append(actor_issue)

        return actor_issues

    def create(self, *args, model_klass=AbstractModel, **kwargs):

        model = model_klass(*args, **kwargs)

        filtered_actors = self.filter_actors()
        filtered_issues = self.filter_issues()

        filtered_actor_issues = self.filter_actor_issues(filtered_actors, filtered_issues)

        for actor in filtered_actors.values():
            model.add_actor(actor_name=actor.fullname, actor_id=actor.id)

        for issue in filtered_issues.values():
            model_issue = model.add_issue(issue.name, issue.description)
            model_issue.lower = issue.lower
            model_issue.upper = issue.upper

            model_issue.calculate_delta()
            model_issue.calculate_step_size()

        for actor_issue in filtered_actor_issues:
            model.add_actor_issue(
                actor=actor_issue.actor,
                issue=actor_issue.issue,
                position=actor_issue.position,
                salience=actor_issue.salience,
                power=actor_issue.power
            )

        return model

    def __call__(self, model_klass, *args, **kwargs):
        return self.create(*args, model_klass=model_klass, **kwargs)



from decide.qt.mainwindow.gui import DecideMainWindow
from decide.qt.mainwindow.settings import SettingsFormWidget, ProgramSettings


def test_load_settings(qtbot):
    """
    Load settings from ProgramSettings
    """

    program_settings = ProgramSettings()

    main_window = DecideMainWindow()

    salience_weight = 0.99

    program_settings.salience_weight = salience_weight

    assert program_settings.salience_weight == salience_weight

    widget = SettingsFormWidget(program_settings, main_window)
    widget.load()

    assert widget.salience_weight.value() == salience_weight


def test_model_variations(qtbot):
    program_settings = ProgramSettings()

    program_settings.start = 0.0
    program_settings.stop = 0.0

    assert program_settings.model_variations == ['0.00']

    program_settings.start = 0.0
    program_settings.step = 0.05
    program_settings.stop = 0.10

    assert program_settings.model_variations == ['0.00', '0.05', '0.10']



import os

__version__ = "2019.3.1.3"

decide_base_path = os.path.dirname(os.path.abspath(__file__))

log_filename = os.path.join(decide_base_path, "decide.log")

data_folder = os.path.join(decide_base_path, "..", "data")
input_folder = os.path.join(data_folder, "input")



import os

import pytest

from decide import input_folder
from decide.data.modelfactory import ModelFactory
from decide.data.reader import InputDataFile
from decide.model.equalgain import EqualGainModel


@pytest.fixture
def model():
    date_file = InputDataFile.open(os.path.join(input_folder, 'sample_data.txt'))

    factory = ModelFactory(date_file=date_file)

    return factory(EqualGainModel)


@pytest.fixture
def sample_model():
    date_file = InputDataFile.open(os.path.join(input_folder, 'sample_data.txt'))

    factory = ModelFactory(date_file=date_file)

    return factory(EqualGainModel)



import os
import xml.etree.cElementTree as ET

from PyQt5 import QtCore
from PyQt5 import QtWidgets
from decide import decide_base_path
from decide.cli import float_range


class SettingsFormWidget(QtWidgets.QFormLayout):
    """
    FormLayout containing the different parameters for the model
    """

    def __init__(self, settings, main_window, *args, **kwargs):
        """
        :type settings: decide.qt.mainwindow.settings.ProgramSettings
        :type main_window: decide.qt.mainwindow.gui.DecideMainWindow
        """
        super().__init__(*args, **kwargs)

        self.is_loading = False

        self.settings = settings
        self.main_window = main_window

        self.fixed_weight = QtWidgets.QDoubleSpinBox()
        self.fixed_weight.setSingleStep(0.05)
        self.fixed_weight.setDecimals(2)
        self.fixed_weight.setValue(0.10)
        self.fixed_weight.setMaximum(1)

        self.salience_weight = QtWidgets.QDoubleSpinBox()
        self.salience_weight.setSingleStep(0.05)
        self.salience_weight.setMaximum(1)
        self.salience_weight.setDecimals(2)
        self.salience_weight.setValue(0.40)

        self.randomized_value = QtWidgets.QDoubleSpinBox()
        self.randomized_value.setSingleStep(0.05)

        self.start = QtWidgets.QDoubleSpinBox()
        self.start.setSingleStep(0.05)
        self.start.valueChanged.connect(self.state_changed)

        self.step = QtWidgets.QDoubleSpinBox()
        self.step.setSingleStep(0.05)
        self.step.valueChanged.connect(self.state_changed)

        self.stop = QtWidgets.QDoubleSpinBox()
        self.stop.setSingleStep(0.05)
        self.stop.valueChanged.connect(self.state_changed)

        self.iterations = QtWidgets.QSpinBox()
        self.iterations.setMinimum(1)
        self.iterations.setValue(10)
        self.iterations.setMaximum(10000)

        self.repetitions = QtWidgets.QSpinBox()
        self.repetitions.setMinimum(1)
        self.repetitions.setValue(10)
        self.repetitions.setMaximum(10000)

        self.addRow(QtWidgets.QLabel("Fixed weight"), self.fixed_weight)
        self.addRow(QtWidgets.QLabel("Salience weight"), self.salience_weight)
        self.addRow(QtWidgets.QLabel(""))
        self.addRow(QtWidgets.QLabel("Negotiation rounds"), self.iterations)
        self.addRow(QtWidgets.QLabel("Simulation repetitions"), self.repetitions)
        self.addRow(QtWidgets.QLabel(""))

        self.addRow(QtWidgets.QLabel("p-value"))
        self.addRow(QtWidgets.QLabel("Start"), self.start)
        self.addRow(QtWidgets.QLabel("Step"), self.step)
        self.addRow(QtWidgets.QLabel("Stop"), self.stop)

    def load(self):
        """
        Copy the values from the ProgramSettings object
        """

        settings = self.settings.__dict__.items()

        self.is_loading = True

        for key, value in settings:

            if hasattr(self, key):
                attr = getattr(self, key)

                if isinstance(value, bool):  # type: QtWidgets.QAction
                    attr.setChecked(value)
                if isinstance(attr, list):
                    self.settings.__dict__[attr] = value
                else:
                    attr.setValue(value)

        self.is_loading = False

    def save(self):
        """
        Set the attributes to the ProgramSettings object
        """
        settings = self.__dict__.items()

        for key, value in settings:
            if hasattr(self.settings, key):
                setattr(self.settings, key, value.value())

    def state_changed(self):

        if not self.is_loading:
            self.save()
        self.main_window.overview_widget.update_widget()


class ProgramSettings(QtCore.QObject):
    """
    The settings for the model parameters
    """

    changed = QtCore.pyqtSignal()

    settings_file = os.path.join(decide_base_path, "decide-settings.xml")

    def __init__(self, *args, **kwargs):
        super(ProgramSettings, self).__init__(*args, **kwargs)

        self.input_filename = ""
        self.output_directory = ""

        self.salience_weight = 0.4
        self.fixed_weight = 0.1

        self.randomized_value = 0.00
        self.start = 0.00
        self.step = 0.05
        self.stop = 0.50

        self.repetitions = 10
        self.iterations = 10

        self.settings_type = "xml"
        self.settings_list_separator = ";"

        self.version = 1

        self.output_sqlite = True
        self.issue_development_csv = True
        self.externalities_csv = True
        self.exchanges_csv = True
        self.voting_positions = True
        self.summary_only = False

        self.selected_actors = []
        self.selected_issues = []

        self.recently_opened = []

    def settings_file_path(self):
        file_path = os.path.join(decide_base_path, self.settings_file)

        return file_path

    def save(self):
        if self.settings_type == "xml":
            self._save_xml()

    def set_input_filename(self, value):
        if value in self.recently_opened:
            self.recently_opened.remove(value)

        self.recently_opened.insert(0, value)

        self.input_filename = value

    def _save_xml(self):

        file_path = self.settings_file_path()

        element = ET.Element("decide-settings")

        for key, value in self.__dict__.items():
            if not key.startswith("_") and hasattr(self, key):

                if isinstance(value, list):
                    value = self.settings_list_separator.join(value)

                child = ET.Element(key)
                child.text = str(value)
                element.append(child)

        ET.ElementTree(element).write(file_path)

    def load(self):
        if self.settings_type == "xml":
            self._load_xml()

    def _load_xml(self):

        file_path = os.path.join(decide_base_path, self.settings_file)

        if not os.path.exists(file_path):
            return

        for elm in ET.parse(file_path).getroot():

            if hasattr(self, elm.tag):

                attr = getattr(self, elm.tag)

                if isinstance(attr, bool):
                    setattr(self, elm.tag, elm.text == "True")
                elif isinstance(attr, int):
                    setattr(self, elm.tag, int(elm.text))
                elif isinstance(attr, float):
                    setattr(self, elm.tag, float(elm.text))
                elif isinstance(attr, list):

                    values = str(elm.text).split(self.settings_list_separator)

                    values = [x.strip() for x in values]

                    setattr(
                        self, elm.tag, values
                    )
                else:
                    setattr(self, elm.tag, str(elm.text))

    @property
    def data_set_name(self):
        return os.path.splitext(os.path.basename(self.input_filename))[0]

    @property
    def model_variations(self):
        """
        Takes start, step and stop and creates a list of values
        :rtype: list of str
        """

        if self.start and self.stop and self.start == self.stop:
            return ['{:.2f}'.format(self.start)]
        else:
            values = [
                '{:.2f}'.format(round(p, 2))
                for p in float_range(
                    start=self.start, step=self.step, stop=self.stop
                )
            ]

            if len(values) > 0:
                return values
            else:
                return ['0.0']



import os

import numpy as np
import pandas as pd
from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param


def write_summary_result(conn, model_run_ids, output_directory):
    df = pd.read_sql("""
    SELECT
      a.p as p,
      a.issue as issue,
      a.iteration as iteration,
      a.repetion as repetion,
      a.numerator / a.denominator AS nbs
    FROM (SELECT
            sum(ai.position * ai.power * ai.salience) AS numerator,
            sum(ai.salience * ai.power)               AS denominator,
            r.pointer                                 AS repetion,
            i2.pointer                                AS iteration,
            m.p,
      i.name as issue
          FROM actorissue ai
            LEFT JOIN issue i ON ai.issue_id = i.id
            LEFT JOIN actor a ON ai.actor_id = a.id
            LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
            LEFT JOIN repetition r ON i2.repetition_id = r.id
            LEFT JOIN modelrun m ON r.model_run_id = m.id        
          WHERE  ai.type = 'before' AND m.id IN (%s)
         GROUP BY m.id,r.id, i2.id, i.id) a
    """ % list_to_sql_param(model_run_ids),
                     conn,
                     index_col='p',
                     columns=['nbs']
                     )

    table_avg = pd.pivot_table(df, index=['issue', 'p'], columns=['iteration'], values=['nbs'], aggfunc=np.average)
    table_var = pd.pivot_table(df, index=['issue', 'p'], columns=['iteration'], values=['nbs'], aggfunc=np.var)

    table_avg.to_csv(os.path.join(output_directory, 'nbs_average.csv'))
    table_var.to_csv(os.path.join(output_directory, 'nbs_variance.csv'))

    sql_2 = """SELECT issue.name, issue.id
FROM issue
INNER JOIN dataset d on issue.data_set_id = d.id
INNER JOIN modelrun m on d.id = m.data_set_id
WHERE m.id IN (%s)
GROUP BY issue.name, issue.id
ORDER BY issue.name""" % list_to_sql_param(model_run_ids)

    cursor = conn.execute_sql(sql=sql_2, params=[])
    issues = cursor.fetchall()

    # %%

    for name, issue_id in issues:
        df = pd.read_sql("""SELECT a.p                         as p,
       a.issue                     as issue,
       a.iteration                 as iteration,
       a.repetion                  as repetion,
       a.numerator / a.denominator AS nbs
FROM (SELECT sum(ai.position * ai.power * ai.salience) AS numerator,
             sum(ai.salience * ai.power)
                                                       AS denominator,
             r.pointer
                                                       AS repetion,
             i2.pointer
                                                       AS iteration,
             m.p,
             i.name                                    as issue
      FROM actorissue ai
               LEFT JOIN issue i ON ai.issue_id = i.id
               LEFT JOIN actor a ON ai.actor_id = a.id
               LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
               LEFT JOIN repetition r ON i2.repetition_id = r.id
               LEFT JOIN modelrun m ON r.model_run_id = m.id
               LEFT JOIN dataset d ON a.data_set_id = d.id
      WHERE ai.type = 'before'
        AND m.id IN(%s)
        AND i.id = %s
      GROUP BY m.id, r.id, i2.id, i.id) a
        """ % (list_to_sql_param(model_run_ids), issue_id),
                         conn,
                         index_col='p',
                         columns=['nbs']
                         )
        table = pd.pivot_table(df, index=['iteration'], columns=['p'], values=['nbs'])
        plt = table.plot()

        plt.set_title(name)
        plt.set_ylim(0, 110)

        lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

        plt.figure.savefig(
            os.path.join(output_directory, 'nbs_{}.png'.format(name)),
            bbox_extra_artists=(lgd,),
            bbox_inches="tight",
        )


if __name__ == '__main__':
    m = Manager(os.environ.get('DATABASE_URL'))
    m.init_database()

    model_run_ids = [43, 44]

    write_summary_result(connection, model_run_ids, data_folder)



import decimal
import logging
import random
from operator import attrgetter

from decide.model import base, calculations


class EqualGainExchangeActor(base.AbstractExchangeActor):
    """
    AbstractExchangeActor is the same actor...
    """

    def __init__(
        self,
        model: "EqualGainModel",
        actor: base.Actor,
        demand_issue: base.Issue,
        supply_issue: base.Issue,
        exchange: "EqualGainExchange",
    ):
        super().__init__(model, actor, demand_issue, supply_issue, exchange)

        self.equal_gain_voting = "-"
        self.z = "-"
        self.u = "-"
        self.v = "-"
        self.eu_max = "-"

    def randomized_gain(self, u, v, z, exchange_ratio, exchange_ratio_j):

        p = self.exchange.model.randomized_value

        eu = self.exchange.gain

        self.z = z
        self.u = u
        self.v = v

        self.opposite_actor.u = u
        self.opposite_actor.v = v
        self.opposite_actor.z = z

        exchange_ratio_zero_i = calculations.exchange_ratio_by_expected_utility(
            exchange_ratio, self.supply.salience, self.demand.salience
        )

        self.eu_max = abs(
            calculations.expected_utility(self, exchange_ratio_zero_i, exchange_ratio)
        )

        exchange_ratio_zero_j = calculations.exchange_ratio_by_expected_utility(
            exchange_ratio_j,
            self.opposite_actor.supply.salience,
            self.opposite_actor.demand.salience,
        )

        self.opposite_actor.eu_max = abs(
            calculations.expected_utility(self, exchange_ratio_zero_j, exchange_ratio_j)
        )

        if v < 0.5:  # V < 0.5:
            eui = p * z * (self.eu_max - eu) + eu
        else:
            eui = eu - p * z * eu

        # calculate the expected utility of J
        move_i = abs(self.supply.position - self.opposite_actor.demand.position)

        # supply exchange ratio for i = q, supply for j = p
        exchange_ratio_q = calculations.exchange_ratio(
            move_i,
            self.supply.salience,
            self.supply.power,
            self.model.nbs_denominators[self.supply.issue],
        )

        # supply exchange ratio for j, demand for i
        exchange_ratio_p = (
            eui + exchange_ratio_q * self.supply.salience
        ) / self.demand.salience

        eui_check = abs(
            calculations.expected_utility(self, exchange_ratio_q, exchange_ratio_p)
        )

        if abs(eui_check - eui) > 1e-10:
            raise Exception("This should be equal.")

        move_j = calculations.reverse_move(
            actor_issues=self.opposite_actor.actor_issues(),
            actor=self.opposite_actor,
            exchange_ratio=exchange_ratio_p,
        )

        euj = abs(
            calculations.expected_utility(
                self.opposite_actor, exchange_ratio_p, exchange_ratio_q
            )
        )

        delta_x_j_supply = abs(
            self.opposite_actor.supply.position - self.demand.position
        )

        if abs(move_j) > delta_x_j_supply:
            exchange_ratio_p = calculations.exchange_ratio(
                delta_x_j_supply,
                self.opposite_actor.supply.salience,
                self.opposite_actor.supply.power,
                self.model.nbs_denominators[self.opposite_actor.supply.issue],
            )

            exchange_ratio_q = abs(
                (eui - exchange_ratio_p * self.demand.salience) / self.supply.salience
            )

            move_j = delta_x_j_supply
            move_i = calculations.reverse_move(
                self.actor_issues(), self, exchange_ratio_q
            )

            if abs(move_i) < abs(
                self.supply.position - self.opposite_actor.demand.position
            ):

                euj = abs(
                    calculations.expected_utility(
                        self.opposite_actor, exchange_ratio_p, exchange_ratio_q
                    )
                )
                eui_check = abs(
                    calculations.expected_utility(
                        self, exchange_ratio_q, exchange_ratio_p
                    )
                )

                # check if the shift does not exceed the NBS.
                if self.opposite_actor.supply.position > self.demand.position:
                    move_j_a = move_j * -1
                else:
                    move_j_a = move_j

                nbs_adjusted = self.opposite_actor.adjust_nbs(
                    position=self.opposite_actor.supply.position + move_j_a
                )
                self.opposite_actor.nbs_0 = nbs_adjusted

                if (
                    self.demand.position >= self.opposite_actor.nbs_0
                    and self.demand.position >= nbs_adjusted
                ):
                    pass
                elif (
                    self.demand.position <= self.opposite_actor.nbs_0
                    and self.demand.position <= nbs_adjusted
                ):
                    pass
                else:

                    delta = abs(
                        calculations.adjusted_nbs_by_position(
                            actor_issues=self.opposite_actor.actor_issues(),
                            updates=self.opposite_actor.exchange.updates[
                                self.opposite_actor.supply.issue
                            ],
                            actor=self.opposite_actor.actor,
                            x_pos=self.opposite_actor.supply.position,
                            new_nbs=self.demand.position,
                            denominator=self.model.nbs_denominators[
                                self.opposite_actor.supply.issue
                            ],
                        )
                    )

                    exchange_ratio_p = calculations.exchange_ratio(
                        delta_x=delta,
                        salience=self.opposite_actor.supply.salience,
                        power=self.opposite_actor.supply.power,
                        denominator=self.model.nbs_denominators[
                            self.opposite_actor.supply.issue
                        ],
                    )

                    exchange_ratio_q = abs(
                        (eui - exchange_ratio_p * self.demand.salience)
                        / self.supply.salience
                    )

                    move_i_b = calculations.reverse_move(
                        actor_issues=self.actor_issues(),
                        actor=self,
                        exchange_ratio=exchange_ratio_p,
                    )

                    euj = abs(
                        calculations.expected_utility(
                            self.opposite_actor, exchange_ratio_p, exchange_ratio_q
                        )
                    )
                    eui_check = abs(
                        calculations.expected_utility(
                            self, exchange_ratio_q, exchange_ratio_p
                        )
                    )

                    move_i = move_i_b
                    move_j = move_j_a

                    self.opposite_actor.is_adjusted_by_nbs = True

            else:
                logging.info(
                    "SoftFail: when the shift for j is to large by a maximum shift of i, "
                    "the maximum shift of j has to result in a lower shift for i. data="
                    + str(self.exchange)
                )

        else:
            # check if the shift does not exceed the NBS.
            if self.supply.position > self.opposite_actor.demand.position:
                move_i_a = move_i * -1
            else:
                move_i_a = move_i

            nbs_adjusted = self.adjust_nbs(position=self.supply.position + move_i_a)

            # the nbs shifts beyond the position of the demand position

            if (
                self.opposite_actor.demand.position >= self.nbs_0
                and self.opposite_actor.demand.position >= nbs_adjusted
            ):
                pass
            elif (
                self.opposite_actor.demand.position <= self.nbs_0
                and self.opposite_actor.demand.position <= nbs_adjusted
            ):
                pass
            else:

                delta = abs(
                    calculations.adjusted_nbs_by_position(
                        actor_issues=self.actor_issues(),
                        updates=self.exchange.updates[self.supply.issue],
                        actor=self.actor,
                        x_pos=self.supply.position,
                        new_nbs=self.opposite_actor.demand.position,
                        denominator=self.model.nbs_denominators[self.supply.issue],
                    )
                )

                exchange_ratio_q = calculations.exchange_ratio(
                    delta,
                    self.supply.salience,
                    self.supply.power,
                    self.model.nbs_denominators[self.supply.issue],
                )

                exchange_ratio_p = (
                    eui + exchange_ratio_q * self.supply.salience
                ) / self.demand.salience

                move_j_b = calculations.reverse_move(
                    actor_issues=self.opposite_actor.actor_issues(),
                    actor=self.opposite_actor,
                    exchange_ratio=exchange_ratio_p,
                )

                euj = abs(
                    calculations.expected_utility(
                        actor=self.opposite_actor,
                        demand_exchange_ratio=exchange_ratio_p,
                        supply_exchange_ratio=exchange_ratio_q,
                    )
                )

                eui_check = abs(
                    calculations.expected_utility(
                        actor=self,
                        demand_exchange_ratio=exchange_ratio_q,
                        supply_exchange_ratio=exchange_ratio_p,
                    )
                )

                if p == 0 and abs(eui - euj) > 1e-10:
                    raise Exception("Fail: adjusting nbs.")

                move_i = move_i_a
                move_j = move_j_b

                self.is_adjusted_by_nbs = True

        if abs(eui - eui_check) > 1e-10:
            logging.info("fail")
            # raise Exception('Fail: the expected utility of the check ({0})'
            #                 ' does not match the expected utility ({1}), {2}.'
            #                 .format(eui_check, eui,
            #                         self.opposite_actor.is_adjusted_by_nbs))

        if self.eu_max is None:
            # if they are not equal (occurs by p=0.0)
            if not abs(euj - eui) < 1e-10:
                if (euj - eui) < 1e-10:
                    raise Exception(
                        "Fail: when eui_max is none (v=False), euj should be larger in this case. {0}".format(
                            self.opposite_actor.is_adjusted_by_nbs
                        )
                    )

        if p == 0 and abs(eui - euj) > 1e-10:
            raise Exception(
                "Fail: Equal Gain not equal with p=0.0 and nbs_adjusted={0}".format(
                    self.opposite_actor.is_adjusted_by_nbs
                )
            )

        self.eu = eui
        self.opposite_actor.eu = euj

        self.move = move_i
        self.opposite_actor.move = move_j

        if self.supply.position > self.opposite_actor.demand.position:
            self.move *= -1

        if self.opposite_actor.supply.position > self.demand.position:
            self.opposite_actor.move *= -1

        self.moves.pop()
        self.moves.append(self.move)

        self.opposite_actor.moves.pop()
        self.opposite_actor.moves.append(self.opposite_actor.move)

        self.y = self.supply.position + self.move
        self.opposite_actor.y = (
            self.opposite_actor.supply.position + self.opposite_actor.move
        )

        b1 = self.is_move_valid(self.move)
        b2 = self.opposite_actor.is_move_valid(self.opposite_actor.move)

        self.exchange.is_valid = b1 and b2

        self.nbs_0 = self.adjust_nbs(self.supply.position)
        self.nbs_1 = self.adjust_nbs(self.y)

        self.opposite_actor.nbs_0 = self.opposite_actor.adjust_nbs(
            self.opposite_actor.supply.position
        )
        self.opposite_actor.nbs_1 = self.opposite_actor.adjust_nbs(
            self.opposite_actor.y
        )


class EqualGainExchange(base.AbstractExchange):
    actor_class = EqualGainExchangeActor

    def __init__(self, i, j, p, q, m, groups):
        super().__init__(i, j, p, q, m, groups)

    def calculate(self):
        # first we try to move j to the position of i on issue p
        # we start with the calculation for j
        self.dp = calculations.by_absolute_move(self.j.actor_issues(), self.j)
        self.dq = calculations.by_exchange_ratio(self.j, self.dp)

        self.i.move = calculations.reverse_move(self.i.actor_issues(), self.i, self.dq)
        self.j.move = abs(self.i.demand.position - self.j.supply.position)

        # if the move exceeds the interval
        if abs(self.i.move) > abs(self.j.demand.position - self.i.supply.position):
            self.dq = calculations.by_absolute_move(self.i.actor_issues(), self.i)
            self.dp = calculations.by_exchange_ratio(self.i, self.dq)

            self.i.move = abs(self.j.demand.position - self.i.supply.position)
            self.j.move = calculations.reverse_move(
                self.j.actor_issues(), self.j, self.dp
            )

        # determine the direction of both moves
        if self.i.supply.position > self.j.demand.position:
            self.i.move *= -1

        if self.j.supply.position > self.i.demand.position:
            self.j.move *= -1

        # keep the moves in memory so we can check the direction of the actor
        self.i.moves.append(self.i.move)
        self.j.moves.append(self.j.move)

        # adjust the voting positions
        self.i.y = self.i.supply.position + self.i.move
        self.j.y = self.j.supply.position + self.j.move

        self.i.equal_gain_voting = self.i.y
        self.j.equal_gain_voting = self.j.y

        # calculate the utility gains for both actors
        eui = calculations.expected_utility(self.i, self.dq, self.dp)
        euj = calculations.expected_utility(self.j, self.dp, self.dq)

        # since this is the Equal Gain model, the gains should be equal
        if calculations.is_gain_equal(eui, euj):
            self.gain = abs(eui)
            self.i.eu = self.gain
            self.j.eu = self.gain

        b1 = self.i.is_move_valid(self.i.move)
        b2 = self.j.is_move_valid(self.j.move)

        self.is_valid = b1 and b2

        if self.gain < 1e-10:
            self.is_valid = False

        if self.is_valid:
            self.i.check_nbs()
            self.j.check_nbs()

            b1 = self.i.is_move_valid(self.i.move)
            b2 = self.j.is_move_valid(self.j.move)

            self.is_valid = b1 and b2
        else:
            return  # stop if its not valid

        if (
            self.model.randomized_value is not None
            and self.model.randomized_value > 0.0
        ):

            u = random.uniform(0, 1)
            v = random.uniform(0, 1)

            z = decimal.Decimal(random.uniform(0, 1))

            if u < 0.5:
                self.i.randomized_gain(u, v, z, self.dp, self.dq)
            else:
                self.j.randomized_gain(u, v, z, self.dq, self.dp)

    def csv_row(self, head=False):

        if head:
            return [
                # the actors
                "actor_name",  # exchange.i.actor_name,
                "supply",  # exchange.i.supply,
                "power",
                "sal s",
                "sal d",
                "sal s/d",
                "start",
                "move",
                "voting",
                "equal gain voting",
                "demand",
                "u",
                "v",
                "z",
                "max eu",
                "eu",
                "gain p=0",
                "nbs 0",
                "nbs 1",
                "delta nbs",
                "check",
                "",
                "actor_name",  # exchange.i.actor_name,
                "supply",  # exchange.i.supply,
                "power",
                "sal s",
                "sal d",
                "sal s/d",
                "start",
                "move",
                "voting",
                "equal gain voting",
                "demand",
                "u",
                "v",
                "z",
                "max eu",
                "eu",
                "gain p=0",
                "nbs 0",
                "nbs 1",
                "delta nbs",
                "check",
            ]

        exchange = self

        delta_nbs_i = abs(exchange.i.nbs_0 - exchange.i.nbs_1)
        delta_nbs_j = abs(exchange.j.nbs_0 - exchange.j.nbs_1)

        eu_i = abs(
            delta_nbs_i * exchange.i.supply.salience
            - delta_nbs_j * exchange.i.demand.salience
        )
        eu_j = abs(
            delta_nbs_j * exchange.j.supply.salience
            - delta_nbs_i * exchange.j.demand.salience
        )

        check_i = abs(eu_i - exchange.i.eu) < 1e-10
        check_j = abs(eu_j - exchange.j.eu) < 1e-10

        return [
            # the actors
            exchange.i.actor.name,
            exchange.i.supply.issue,
            exchange.i.supply.power,
            exchange.i.supply.salience,
            exchange.i.demand.salience,
            exchange.i.supply.salience / exchange.i.demand.salience,
            exchange.i.supply.position,
            exchange.i.move,
            exchange.i.y,
            exchange.i.equal_gain_voting,
            exchange.i.opposite_actor.demand.position,
            exchange.i.u,
            exchange.i.v,
            exchange.i.z,
            exchange.i.eu_max,
            exchange.i.eu,
            exchange.gain,
            exchange.i.nbs_0,
            exchange.i.nbs_1,
            delta_nbs_i,
            check_i,
            "",
            exchange.j.actor.name,
            exchange.j.supply.issue,
            exchange.j.supply.power,
            exchange.j.supply.salience,
            exchange.j.demand.salience,
            exchange.j.supply.salience / exchange.j.demand.salience,
            exchange.j.supply.position,
            exchange.j.move,
            exchange.j.y,
            exchange.j.equal_gain_voting,
            exchange.j.opposite_actor.demand.position,
            exchange.j.u,
            exchange.j.v,
            exchange.j.z,
            exchange.j.eu_max,
            exchange.j.eu,
            exchange.gain,
            exchange.j.nbs_0,
            exchange.j.nbs_1,
            delta_nbs_j,
            check_j,
        ]


class EqualGainModel(base.AbstractModel):
    """
    Equal Gain implementation
    """

    """
    there can be a random component added to
    the model but this gives not equal outcomes for testing purpose
    """
    ALLOW_RANDOM = True

    def __init__(self, randomized_value=None):
        super().__init__()

        if isinstance(randomized_value, str):
            randomized_value = decimal.Decimal(randomized_value)

        self.randomized_value = randomized_value

        self.model_name = "equal"

        if randomized_value:
            self.model_name += "-" + str(round(randomized_value, 2))

    def sort_exchanges(self):
        """
        The exchanges are sorted by there (equal) gain, highest first
        """
        self.exchanges.sort(key=attrgetter("gain"), reverse=True)

    def highest_gain(self):
        """
        Overrides Abstract # To sort the list in place...
        :return:
        """
        self.sort_exchanges()

        realize = self.exchanges.pop(0)

        if len(self.exchanges) > 0:

            # in some cases the exchanges have an equal gain, choice randomly between them
            if EqualGainModel.ALLOW_RANDOM:

                next_exchange = self.exchanges[0]

                if abs(realize.gain - next_exchange.gain) < 1e-20:
                    self.tie_count += 1

                    if random.random() >= 0.5:
                        self.exchanges.append(realize)
                        realize = self.exchanges.pop(0)
                    else:
                        self.exchanges.append(next_exchange)

        return realize

    @staticmethod
    def new_exchange_factory(i, j, p, q, model, groups):
        return EqualGainExchange(i, j, p, q, model, groups)



import logging
import os
import sys
import time
import xml.etree.cElementTree as ET

from PyQt5 import QtCore
from PyQt5 import QtWidgets
from decide import log_filename
from decide.cli import init_output_directory
from decide.data.modelfactory import ModelFactory
from decide.data.reader import InputDataFile
from decide.model.equalgain import EqualGainModel
from decide.model.observers.exchanges_writer import ExchangesWriter
from decide.model.observers.externalities import Externalities
from decide.model.observers.issue_development import IssueDevelopment
from decide.model.observers.logger import Logger
from decide.model.observers.observer import Observable
from decide.model.observers.sqliteobserver import SQLiteObserver
from decide.model.utils import ModelLoop
from decide.qt import utils
from decide.qt.mainwindow.settings import ProgramSettings
from decide.qt.mainwindow.settings import SettingsFormWidget
from decide.qt.mainwindow.widgets import ActorWidget, IssueWidget, SummaryWidget, ActorIssueWidget, MenuBar
from decide.qt.utils import show_user_error


class Worker(QtCore.QObject):
    """
    Worker to execute the model in a thread so the window does not freeze
    """
    finished = QtCore.pyqtSignal(str)
    update = QtCore.pyqtSignal(int, int, int, float)

    def __init__(self, settings: ProgramSettings):
        super(Worker, self).__init__()

        self.settings = settings
        self.break_loop = False

    @QtCore.pyqtSlot()
    def run_model(self):
        settings = self.settings

        factory = ModelFactory(
            date_file=InputDataFile.open(settings.input_filename),
            actor_whitelist=settings.selected_actors,
            issue_whitelist=settings.selected_issues,
        )

        parent_output_directory = init_output_directory(
            settings.output_directory,
            settings.data_set_name,
        )

        model = factory(model_klass=EqualGainModel, randomized_value=settings.model_variations[0])

        event_handler = init_event_handlers(model, parent_output_directory, settings)

        event_handler.before_model()

        model_variations = list(settings.model_variations)

        for variation, p in enumerate(model_variations, 1):

            output_directory = init_output_directory(
                settings.output_directory,
                settings.data_set_name,
                p
            )

            event_handler.update_output_directory(output_directory)

            event_handler.before_repetitions(
                repetitions=settings.repetitions,
                iterations=settings.iterations,
                randomized_value=p,
            )

            start_time = time.time()

            for repetition in range(settings.repetitions):

                model = factory(model_klass=EqualGainModel, randomized_value=p)

                event_handler.update_model_ref(model)

                model_loop = ModelLoop(model, event_handler, repetition)

                event_handler.before_iterations(repetition)

                for iteration_number in range(settings.iterations):

                    if self.break_loop:
                        break

                    self.update.emit(variation, repetition, iteration_number, start_time)

                    model_loop.loop()

                event_handler.after_iterations(repetition)

                if self.break_loop:
                    break

            if not self.break_loop:
                event_handler.after_repetitions()

                logging.info('tie count is {}'.format(model.tie_count))

        if not self.break_loop:
            event_handler.update_output_directory(parent_output_directory)
            event_handler.after_model()

        self.finished.emit(parent_output_directory)

    def stop(self):
        self.break_loop = True


class ProgramData(QtCore.QObject):
    """
    Central object for the data used for displaying
    """

    changed = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.issues = {}
        self.actors = {}
        self.actor_issues = {}


def init_event_handlers(model, output_directory, settings):
    """
    :type model: decide.model.base.AbstractModel
    :type output_directory: str
    :type settings: ProgramSettings
    """

    event_handler = Observable(model_ref=model, output_directory=output_directory)

    Logger(event_handler)

    SQLiteObserver(event_handler, settings.output_directory)
    Externalities(event_handler, False)
    ExchangesWriter(event_handler, False)
    IssueDevelopment(event_handler, False)

    return event_handler


class DecideMainWindow(QtWidgets.QMainWindow):
    """
    The first thing you see
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data = ProgramData()
        self.settings = ProgramSettings()
        self.settings.load()
        self.menu_bar = MenuBar(self, self)
        self.start = QtWidgets.QPushButton("Start")

        self.issue_widget = IssueWidget(self)
        self.actor_widget = ActorWidget(self)
        self.scroll_area_widget = QtWidgets.QWidget()
        self.actor_issue_widget = ActorIssueWidget(self.scroll_area_widget)
        self.scroll_area_widget.setLayout(self.actor_issue_widget)
        self.settings_widget = SettingsFormWidget(self.settings, self)

        self.progress_dialog = None

        self.thread = QtCore.QThread()
        self.worker = None

        self.overview_widget = SummaryWidget(
            self, self.settings, self.data, self.actor_widget, self.issue_widget
        )

        self.init_ui()

        self.load_settings()

        self.init_ui_data(self.settings.input_filename)

    def init_ui(self):
        self.statusBar().showMessage("Ready")

        self.setMenuBar(self.menu_bar)

        self.menu_bar.setNativeMenuBar(False)

        main = QtWidgets.QHBoxLayout()

        self.start.clicked.connect(self.run)
        self.set_start_button_state()

        left = QtWidgets.QVBoxLayout()

        left_top = QtWidgets.QHBoxLayout()

        actor_box = QtWidgets.QGroupBox("Actors")
        actor_box.setLayout(self.actor_widget)

        issue_box = QtWidgets.QGroupBox("Issues")
        issue_box.setLayout(self.issue_widget)

        # for the ActorIssueBox
        scroll_area = QtWidgets.QScrollArea(self)
        scroll_area.setWidgetResizable(True)

        scroll_area.setWidget(self.scroll_area_widget)

        scroll_area_layout = QtWidgets.QHBoxLayout()
        scroll_area_layout.addWidget(scroll_area)
        # the grid layout = self.actor_issue_widget

        actor_issue_box = QtWidgets.QGroupBox("Actor Issues")
        actor_issue_box.setLayout(scroll_area_layout)

        left_top.addWidget(actor_box, 1)
        left_top.addWidget(issue_box, 1)

        left.addLayout(left_top, 1)
        left.addWidget(actor_issue_box, 1)

        settings_box = QtWidgets.QGroupBox("Model parameters")
        settings_box.setLayout(self.settings_widget)

        overview_box = QtWidgets.QGroupBox("Overview")
        overview_box.setLayout(self.overview_widget)
        self.overview_widget.setAlignment(QtCore.Qt.AlignTop)

        right = QtWidgets.QVBoxLayout()
        right.addWidget(settings_box, 1)
        right.addWidget(overview_box, 1)
        right.addWidget(self.start, 1)
        right.setAlignment(QtCore.Qt.AlignTop)

        main.addLayout(left, 1)
        main.addLayout(right, 1)

        q = QtWidgets.QWidget()
        q.setLayout(main)

        self.setCentralWidget(q)

        self.setGeometry(300, 300, 1024, 768)
        self.setWindowTitle("Decide Exchange Model")
        self.show()

    def show_debug_dialog(self):
        utils.open_file_natively(log_filename)

    def show_error_report_dialog(self):

        from decide.qt.mainwindow.errordialog import ErrorDialog
        ex = ErrorDialog("Error message", self)

    def update_data_widgets(self):

        self.issue_widget.set_issues(list(self.data.issues.values()))
        self.actor_widget.set_actors(list(self.data.actors.values()))

        self.actor_issue_widget.set_actor_issues(list(self.data.actor_issues.values()))

        self.overview_widget.update_widget()

    def update_actor_issue_widget(self):
        """
        Button event to update the ActorIssueWidget
        """
        issue = self.data.issues[self.sender().objectName()]
        actor_issues = list(self.data.actor_issues[issue].values())

        self.actor_issue_widget.set_actor_issues(issue, actor_issues)
        self.overview_widget.update_widget()

    def open_input_data(self):
        """
        Open the file, parse the data and update the layout
        """
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Select input data")

        if file_name:
            self.load_input_data(file_name)

    def load_input_data(self, file_name):

        self.statusBar().showMessage("Input file set to {} ".format(file_name))

        self.init_ui_data(file_name)

        self.overview_widget.update_widget()

    def open_current_input_window_with_current_data(self):

        from decide.qt.inputwindow.gui import InputWindow, register_app

        ex = InputWindow(self)
        ex.load(self.settings.input_filename)

        register_app(ex)

    def select_output_dir(self):
        """
        Output directory
        """
        self.settings.output_directory = str(
            QtWidgets.QFileDialog.getExistingDirectory(self, "Select Directory")
        )
        self.statusBar().showMessage(
            "Output directory set to: {} ".format(self.settings.output_directory)
        )

        self.overview_widget.update_widget()

    def set_start_button_state(self):
        if (
                self.settings.input_filename
                and self.settings.output_directory
                and not self.thread.isRunning()
        ):
            self.start.setEnabled(True)
            self.start.setDisabled(False)
        else:
            self.start.setEnabled(False)
            self.start.setDisabled(True)

    def open_data_view(self):

        from decide.qt.inputwindow.gui import InputWindow, register_app

        ex = InputWindow(self)

        register_app(ex)

    def init_ui_data(self, file_name):

        if file_name:
            if not os.path.isfile(file_name):
                show_user_error(self, "Selected file does not exists")
            else:
                data_file = InputDataFile.open(file_name)

                if data_file.is_valid:

                    self.data.actors = data_file.actors
                    self.data.issues = data_file.issues
                    self.data.actor_issues = data_file.actor_issues
                    self.update_data_widgets()

                    # update the settings
                    self.settings.set_input_filename(file_name)
                    self.settings.save()
                else:
                    show_user_error(
                        self,
                        "Your input file contains errors. The red rows are not valid. Hover with your move over the row for more information.",
                    )
                    from decide.qt.mainwindow.errorgrid import ErrorGrid

                    ex = ErrorGrid(data_file, self)

    def run_safe(self):

        self.worker = Worker(self.settings)
        self.worker.finished.connect(self.finished)
        self.worker.moveToThread(self.thread)
        self.worker.finished.connect(self.thread.quit)
        self.worker.update.connect(self.update_progress)
        self.thread.started.connect(self.worker.run_model)
        self.thread.start()

    def update_progress(self, variation, repetition, iteration, start_time):

        value = variation * repetition * self.settings.iterations + iteration

        self.progress_dialog.setValue(value)
        repetitions = self.settings.repetitions

        if repetition > 0 and repetitions > 100 and repetition % 20 == 0:
            time_expired = time.time() - start_time

            avg_time_per_repetition = time_expired / repetition

            estimated_time_needed_seconds = avg_time_per_repetition * repetitions

            estimated_time_left = estimated_time_needed_seconds - (
                    repetition * avg_time_per_repetition
            )

            self.statusBar().showMessage(
                "{:.0f} minutes remaining ".format(estimated_time_left / 60)
            )

    def finished(self, output_directory):

        self._clean_progress_dialog()

        if not self.worker.break_loop:
            button_reply = QtWidgets.QMessageBox.question(
                self,
                "Done",
                "Done running the calculations. Open the result directory?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )

            if button_reply == QtWidgets.QMessageBox.Yes:
                utils.open_file_natively(output_directory)

    def _clean_progress_dialog(self):
        self.progress_dialog.setValue(
            self.settings.iterations * self.settings.repetitions * len(self.settings.model_variations)
        )
        self.progress_dialog.hide()

    def run(self):
        # store the current state of the app
        self.save_settings()

        self.setWindowTitle("Decide Exchange Model")

        self.show_progress_dialog()

        self.run_safe()

    def show_progress_dialog(self):

        self.progress_dialog = QtWidgets.QProgressDialog(
            "Task in progress",
            "Cancel",
            0,
            self.settings.repetitions * self.settings.iterations * len(self.settings.model_variations),
            self,
        )

        self.progress_dialog.setWindowTitle('Progress')

        bar = QtWidgets.QProgressBar(self.progress_dialog)
        bar.setTextVisible(True)
        bar.setFormat("%v/%m (%p%)")
        bar.setMinimum(0)
        bar.setMaximum(self.settings.iterations * self.settings.repetitions * len(self.settings.model_variations))
        self.progress_dialog.setBar(bar)

        self.progress_dialog.canceled.connect(self.cancel)

        self.progress_dialog.show()

    def load_settings(self):

        if os.path.isfile(self.settings.settings_file):
            try:
                self.settings.load()
                self.menu_bar.load()
                self.settings_widget.load()

            except ET.ParseError:
                logging.info("Invalid xml")

    def save_settings(self):
        self.settings_widget.save()
        self.issue_widget.save()
        self.actor_widget.save()
        self.menu_bar.save()
        self.settings.save()

    def cancel(self):
        self.worker.stop()


def main():
    logging.basicConfig(
        filename=log_filename,
        filemode="a",
        level=logging.INFO,
        format=" %(asctime)s - %(levelname)s - %(message)s",
    )

    qtapp = QtWidgets.QApplication(sys.argv)
    qtapp.setQuitOnLastWindowClosed(True)

    app = DecideMainWindow()

    sys.excepthook = utils.exception_hook

    sys.exit(qtapp.exec_())


if __name__ == "__main__":
    main()



import logging
import sys

import requests
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QMessageBox, QDialog

from decide import log_filename
from decide.qt.mainwindow.settings import ProgramSettings
from decide.qt.utils import exception_hook


class ErrorDialog(QDialog):
    """
    Dialog to send error information to me
    """
    def __init__(self, message="", *args, **kwargs):
        super(ErrorDialog, self).__init__(*args, **kwargs)

        self.settings = ProgramSettings()
        self.settings.load()

        self.main = QtWidgets.QVBoxLayout()

        self.error_message = self.add_message_box(message)

        self.add_label("Attachments:")
        self.settings_checkbox = self.add_settings()
        self.input_file_checkbox = self.add_input_file()
        self.add_send_button()

        self.init_window()

    def init_window(self):

        central_widget = QtWidgets.QWidget()
        central_widget.setLayout(self.main)
        self.setLayout(self.main)

        self.setGeometry(300, 300, 400, 400)
        self.setWindowTitle("Error reporting tool")
        self.show()

    def add_widget(self, widget):
        self.main.addWidget(widget)

    def add_label(self, text):
        self.add_widget(QtWidgets.QLabel(text))

    def add_message_box(self, message):

        self.add_widget(QtWidgets.QLabel("What happened?"))

        textarea = QtWidgets.QPlainTextEdit()
        textarea.setPlainText(message)

        self.add_widget(textarea)

        return textarea

    def add_input_file(self):

        checkbox = QtWidgets.QCheckBox(self.settings.input_filename)
        checkbox.setChecked(True)
        self.add_widget(checkbox)

        return checkbox

    def add_settings(self):

        checkbox = QtWidgets.QCheckBox(self.settings.settings_file)
        checkbox.setChecked(True)
        self.add_widget(checkbox)

        return checkbox

    def add_send_button(self):

        button = QtWidgets.QPushButton()
        button.setText("Send error report")
        button.clicked.connect(self.send_error_report)
        self.add_widget(button)

        return button

    def send_error_report(self):

        confirm = QtWidgets.QMessageBox.question(
            self,
            "Confirm sending",
            "Are you sure you want to send the message with the attachments?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        settings = None
        input_data = None

        if confirm == QMessageBox.Yes:

            files = {}

            if self.settings_checkbox.isChecked():
                settings = open(self.settings.settings_file_path())
                files["settings"] = settings
            if self.input_file_checkbox.isChecked():
                input_data = open(self.settings.input_filename)
                files["input_data"] = input_data

            response = requests.post(
                url="https://jelmertmail.nl/decide/",
                data=dict(message=self.error_message.toPlainText()),
                files=files,
            )

            QtWidgets.QMessageBox.about(self, "Server response", str(response.content))

            if settings:
                settings.close()

            if input_data:
                input_data.close()


def main():

    logging.basicConfig(
        filename=log_filename,
        filemode="w",
        level=logging.DEBUG,
        format=" %(asctime)s - %(levelname)s - %(message)s",
    )

    sys.excepthook = exception_hook

    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)

    error_dialog = ErrorDialog("empty message")

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import csv
from copy import copy
from typing import List, Dict

import typesystem

from .types import PartialActor, ActorIssue, PartialIssue, IssuePosition, IssueDescription, Comment

types = {
    PartialActor.starts_with: PartialActor,
    ActorIssue.starts_with: ActorIssue,
    PartialIssue.starts_with: PartialIssue,
    IssuePosition.starts_with: IssuePosition,
    IssueDescription.starts_with: IssueDescription,
    Comment.starts_with: Comment,
}


class InputDataFile:
    def __init__(self):
        self.errors = {}
        self.rows = {}
        self.data = {}

    def add_typed_object(self, obj):

        klass = obj.__class__

        if klass in self.data:
            self.data[klass][obj] = obj
        else:
            self.data[klass] = {obj: obj}

    @classmethod
    def open(cls, filename: str) -> "InputDataFile":
        """
        Transforms a file with comma separated values to a dictionary where the key is the row number
        """

        data = cls()

        with open(filename, "rt", encoding="utf-8", errors="replace") as csv_file:
            # guess the document format
            dialect = csv.Sniffer().sniff(csv_file.read(1024))
            csv_file.seek(0)

            reader = csv.reader(csv_file, dialect=dialect)

            data.parse_rows(reader)
            data.update_issues_with_positions()
            data.validate_actor_issue_positions()

        return data

    def parse_rows(self, items):
        for index, row in enumerate(items):

            # keep the original data
            self.rows[index] = row

            try:
                type_obj = csv_row_to_type(row)
                self.add_typed_object(type_obj)
            except typesystem.ValidationError as e:
                self.errors[index] = e  # collect the error for displaying purpose

    @property
    def is_valid(self):
        return len(self.errors) == 0

    @property
    def actors(self) -> Dict[str, PartialActor]:
        return self.data[PartialActor]

    @property
    def issues(self) -> Dict[str, PartialIssue]:
        return self.data[PartialIssue]

    @property
    def actor_issues(self) -> Dict[str, ActorIssue]:
        return self.data[ActorIssue]

    @property
    def issue_positions(self) -> Dict[str, IssuePosition]:
        return self.data[IssuePosition]

    def update_issues_with_positions(self):
        """
        Once the file is complete, we can update the lower an upper positions of the issue
        """

        if IssuePosition in self.data:

            for issue_position in self.issue_positions.values():

                if issue_position.issue in self.issues:

                    issue = self.issues[issue_position.issue]

                    if issue.lower is None:
                        issue.lower = issue_position.position
                    elif issue_position.position < issue.lower:
                        issue.lower = issue_position

                    if issue.upper is None:
                        issue.upper = issue_position.position
                    elif issue_position.position > issue.upper:
                        issue.upper = issue_position.position

        self.set_default_issue_positions()

    def set_default_issue_positions(self):
        for issue in self.issues.values():

            if issue.lower is None:
                issue.lower = 0

            if issue.upper is None:
                issue.upper = 100

    def validate_actor_issue_positions(self):
        """
        Validate the positions of the actor issues against the lower & upper issue bounds
        """

        # find the starting position of the actor issues, so we can show the error at the correct position
        row_index_correction = 0

        for type_class in types.values():
            if type_class in self.data and type_class != ActorIssue:
                row_index_correction += len(self.data[type_class])

        for index, actor_issue in enumerate(self.actor_issues.values(), row_index_correction + 1):

            if actor_issue.actor not in self.actors:
                self.errors[index] = typesystem.ValidationError(
                    key='actor',
                    text='{} not found in document'.format(actor_issue.actor)
                )

            if actor_issue.issue in self.issues:
                issue = self.issues[actor_issue.issue]

                try:
                    actor_issue.validate_position(issue)
                except typesystem.ValidationError as e:

                    if index in self.errors:
                        self.errors[index] = e
                    else:
                        self.errors[index] = e
            else:
                self.errors[index] = typesystem.ValidationError(
                    key='issue',
                    text='{} not found document'.format(actor_issue.issue)
                )


def csv_row_to_type(row: List[str]):
    """
    Translate a list of values to the corresponding object
    """
    key = row[0]  # the first element contains the #id field
    row = row[1:]  # the rest the row

    if key not in types:
        raise Exception(f"Add key {key} to Reader.types (row row: {row}")

    row_type = types[key]

    field_names = row_type.fields.keys()

    row = squash(len(row_type.fields), row)

    obj = row_type.validate(dict(zip(field_names, row)))

    return obj


def squash(fields: int, data: List[str], delimiter=" ") -> List[str]:
    """
    Finds out how many fields there are and joins the overhead in to the lasted field

    i.e:
        The object x, y, z contains 3 field.
        The row x,y,z,a,b has 5 values.
        The values a & b will be squashed to z with the given delimiter
    """

    if fields >= len(data):
        return data

    output = copy(data)
    del output[-1]
    output[-1] = delimiter.join(data[fields - 1:])

    return output



import pytest


def test_float_range():
    from decide.cli import float_range

    xrange = list(float_range(0, 0.20, 0.05))

    assert sum([0.0, 0.05, 0.10, 0.15, 0.20]) == sum(xrange)

    assert 0 == sum(float_range(0, 0, 0))

    values = list(float_range(0.05, 0.1, 0.05))

    assert 0.15 == round(sum(values), 2), sum(values)

    assert 0 == sum(float_range(0, 0, 0.10))

    assert 0 == sum(float_range(0, 1, 0.0))

    with pytest.raises(RuntimeError):
        assert 0 == sum(float_range(0, 100, 0.00005)), "Should trigger RuntimeError"

    assert 0 == sum(float_range(0, 1, 0.0))



from decimal import Decimal

from PyQt5 import QtWidgets


def clear_layout(layout):
    """
    Clear al the widgets recursively
    """
    if layout is not None:
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            else:
                clear_layout(item.layout())


class DoubleInput(QtWidgets.QLineEdit):
    """
    Helper for double inputs
    """
    def __init__(self):
        super(DoubleInput, self).__init__()

        self.setFixedWidth(75)

    def setValue(self, value: Decimal):
        if isinstance(value, Decimal):
            value = str(float(value.normalize()))
        if isinstance(value, int):
            value = str(value)
        if isinstance(value, float):
            value = str(value)

        self.setText(value)

    @property
    def valueChanged(self):
        return self.textChanged


def esc(value):
    """
    Escape function used for writing csv files with strings
    """
    return "'{}'".format(value)


def normalize(value):
    """
    Normalize a decimal or float to string, used for writing csv files and the input window
    """

    if isinstance(value, Decimal):
        # float converts exponential values to readable values
        return str(float(value.normalize()))

    if isinstance(value, float):
        return str(value)

    return str(value)




from unittest import TestCase

from decide.model.base import Issue


class TestIssue(TestCase):
    def test_normalization(self):

        # the normal case, works always
        x = Issue(name="test", lower=0, upper=100)

        self.assertEqual(x.step_size, 1)
        self.assertEqual(x.de_normalize(50), 50)

        # case with a smaller delta
        y = Issue(name="test", lower=0, upper=60)

        self.assertEqual(y.step_size, 100 / 60)
        self.assertAlmostEqual(y.de_normalize(50), 30, delta=0.00001)

        # case with a negative delta
        z = Issue(name="test", lower=0, upper=-60)

        self.assertEqual(z.step_size, 100 / -60)
        self.assertAlmostEqual(z.de_normalize(50), -30, delta=1e-10)
        self.assertAlmostEqual(z.de_normalize(0), 0, delta=1e-10)
        self.assertAlmostEqual(z.de_normalize(100), -60, delta=1e-10)

        # case with a different starting position

        a = Issue(name="test", lower=10, upper=90)

        self.assertEqual(a.step_size, 100 / 80)
        self.assertEqual(a.delta, 80)
        self.assertEqual(a.de_normalize(50), 50)
        self.assertEqual(a.de_normalize(0), 10)

        # case with a different starting position and negative delta

        b = Issue(name="test", lower=-10, upper=-90)

        self.assertEqual(b.step_size, 100 / -80)
        self.assertEqual(b.delta, -80)
        self.assertEqual(b.de_normalize(50), -50)
        self.assertEqual(b.de_normalize(0), -10)

        # negative delta, positive lower point, negative upper

        c = Issue(name="test", lower=10, upper=-90)

        self.assertEqual(c.step_size, 100 / -100)
        self.assertEqual(c.delta, -100)
        self.assertEqual(c.de_normalize(50), -40)
        self.assertEqual(c.de_normalize(0), 10)

    def test___eq__(self):

        issue = Issue(name="test")
        issue2 = Issue(name="test")
        issue3 = Issue(name="test3")

        self.assertTrue(issue == issue)
        self.assertTrue(issue == issue2)
        self.assertTrue(issue != issue3)
